@@ -1,23 +1,21 @@
 /*
- * This file is part of oFono - Open Source Telephony
  *
- * Copyright (C) 2009 Nokia Corporation and/or its subsidiary(-ies).
+ *  oFono - Open Source Telephony
  *
- * Contact: Aki Niemi <aki.niemi@nokia.com>
+ *  Copyright (C) 2009-2010 Nokia Corporation and/or its subsidiary(-ies).
  *
- * This program is free software; you can redistribute it and/or
- * modify it under the terms of the GNU General Public License
- * version 2 as published by the Free Software Foundation.
+ *  This program is free software; you can redistribute it and/or modify
+ *  it under the terms of the GNU General Public License version 2 as
+ *  published by the Free Software Foundation.
  *
- * This program is distributed in the hope that it will be useful, but
- * WITHOUT ANY WARRANTY; without even the implied warranty of
- * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
- * General Public License for more details.
+ *  This program is distributed in the hope that it will be useful,
+ *  but WITHOUT ANY WARRANTY; without even the implied warranty of
+ *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+ *  GNU General Public License for more details.
  *
- * You should have received a copy of the GNU General Public License
- * along with this program; if not, write to the Free Software
- * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
- * 02110-1301 USA
+ *  You should have received a copy of the GNU General Public License
+ *  along with this program; if not, write to the Free Software
+ *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
  *
  */
 
@@ -133,8 +131,8 @@ static gboolean decode_reg_status(struct netreg_data *nd, const guint8 *msg,
 				!g_isi_sb_iter_get_byte(&iter, &hsupa, 21))
 				return FALSE;
 
-			*ci = word;
-			*lac = dword;
+			*ci = (int)dword;
+			*lac = (int)word;
 
 			switch (nd->rat) {
 
@@ -181,8 +179,9 @@ static gboolean decode_reg_status(struct netreg_data *nd, const guint8 *msg,
 	return TRUE;
 }
 
-static void reg_status_ind_cb(GIsiClient *client, const void *restrict data,
-				size_t len, uint16_t object, void *opaque)
+static void reg_status_ind_cb(GIsiClient *client,
+				const void *restrict data, size_t len,
+				uint16_t object, void *opaque)
 {
 	const unsigned char *msg = data;
 	struct ofono_netreg *netreg = opaque;
@@ -202,8 +201,9 @@ static void reg_status_ind_cb(GIsiClient *client, const void *restrict data,
 	}
 }
 
-static bool reg_status_resp_cb(GIsiClient *client, const void *restrict data,
-				size_t len, uint16_t object, void *opaque)
+static gboolean reg_status_resp_cb(GIsiClient *client,
+					const void *restrict data, size_t len,
+					uint16_t object, void *opaque)
 {
 	const unsigned char *msg = data;
 	struct isi_cb_data *cbd = opaque;
@@ -222,7 +222,7 @@ static bool reg_status_resp_cb(GIsiClient *client, const void *restrict data,
 	}
 
 	if (len < 3 || msg[0] != NET_REG_STATUS_GET_RESP)
-		goto error;
+		return FALSE;
 
 	if (msg[1] != NET_CAUSE_OK) {
 		DBG("Request failed: %s", net_isi_cause_name(msg[1]));
@@ -241,7 +241,7 @@ error:
 
 out:
 	g_free(cbd);
-	return true;
+	return TRUE;
 }
 
 static void isi_registration_status(struct ofono_netreg *netreg,
@@ -255,7 +255,7 @@ static void isi_registration_status(struct ofono_netreg *netreg,
 		NET_REG_STATUS_GET_REQ
 	};
 
-	if (!cbd)
+	if (!cbd || !nd)
 		goto error;
 
 	if (g_isi_request_make(nd->client, msg, sizeof(msg),
@@ -264,14 +264,13 @@ static void isi_registration_status(struct ofono_netreg *netreg,
 		return;
 
 error:
-	if (cbd)
-		g_free(cbd);
-
 	CALLBACK_WITH_FAILURE(cb, -1, -1, -1, -1, data);
+	g_free(cbd);
 }
 
-static bool name_get_resp_cb(GIsiClient *client, const void *restrict data,
-				size_t len, uint16_t object, void *opaque)
+static gboolean name_get_resp_cb(GIsiClient *client,
+					const void *restrict data, size_t len,
+					uint16_t object, void *opaque)
 {
 	const unsigned char *msg = data;
 	struct isi_cb_data *cbd = opaque;
@@ -290,7 +289,7 @@ static bool name_get_resp_cb(GIsiClient *client, const void *restrict data,
 	}
 
 	if (len < 3 || msg[0] != NET_OPER_NAME_READ_RESP)
-		return false;
+		return FALSE;
 
 	if (msg[1] != NET_CAUSE_OK) {
 		DBG("Request failed: %s", net_isi_cause_name(msg[1]));
@@ -304,7 +303,8 @@ static bool name_get_resp_cb(GIsiClient *client, const void *restrict data,
 		switch (g_isi_sb_iter_get_id(&iter)) {
 
 		case NET_GSM_OPERATOR_INFO:
-			if (!g_isi_sb_iter_get_oper_code(&iter, op.mcc, op.mnc, 2))
+			if (!g_isi_sb_iter_get_oper_code(&iter, op.mcc,
+								op.mnc, 2))
 				goto error;
 			break;
 
@@ -340,7 +340,7 @@ error:
 
 out:
 	g_free(cbd);
-	return true;
+	return TRUE;
 }
 
 
@@ -360,7 +360,7 @@ static void isi_current_operator(struct ofono_netreg *netreg,
 		0x00  /* No sub-blocks */
 	};
 
-	if (!cbd)
+	if (!cbd || !nd)
 		goto error;
 
 	if (g_isi_request_make(nd->client, msg, sizeof(msg),
@@ -374,8 +374,9 @@ error:
 }
 
 
-static bool available_resp_cb(GIsiClient *client, const void *restrict data,
-				size_t len, uint16_t object, void *opaque)
+static gboolean available_resp_cb(GIsiClient *client,
+					const void *restrict data, size_t len,
+					uint16_t object, void *opaque)
 {
 	const unsigned char *msg = data;
 	struct isi_cb_data *cbd = opaque;
@@ -393,7 +394,7 @@ static bool available_resp_cb(GIsiClient *client, const void *restrict data,
 	}
 
 	if (len < 3 || msg[0] != NET_AVAILABLE_GET_RESP)
-		return false;
+		return FALSE;
 
 	if (msg[1] != NET_CAUSE_OK) {
 		DBG("Request failed: %s", net_isi_cause_name(msg[1]));
@@ -404,25 +405,23 @@ static bool available_resp_cb(GIsiClient *client, const void *restrict data,
 	total = msg[2] / 2;
 	list = alloca(total * sizeof(struct ofono_network_operator));
 
-	g_isi_sb_iter_init(&iter, msg, len, 3);
+	for (g_isi_sb_iter_init(&iter, msg, len, 3);
+		g_isi_sb_iter_is_valid(&iter);
+		g_isi_sb_iter_next(&iter)) {
 
-	while (g_isi_sb_iter_is_valid(&iter)) {
+		struct ofono_network_operator *op;
+		char *tag = NULL;
+		guint8 taglen = 0;
+		guint8 status = 0;
+		guint8 umts = 0;
 
 		switch (g_isi_sb_iter_get_id(&iter)) {
 
-		case NET_AVAIL_NETWORK_INFO_COMMON: {
-			struct ofono_network_operator *op;
-			char *tag = NULL;
-			guint8 taglen = 0;
-			guint8 status = 0;
+		case NET_AVAIL_NETWORK_INFO_COMMON:
 
-			if (!g_isi_sb_iter_get_byte(&iter, &status, 2))
-				goto error;
-
-			if (!g_isi_sb_iter_get_byte(&iter, &taglen, 5))
-				goto error;
-
-			if (!g_isi_sb_iter_get_alpha_tag(&iter, &tag,
+			if (!g_isi_sb_iter_get_byte(&iter, &status, 2)
+				|| !g_isi_sb_iter_get_byte(&iter, &taglen, 5)
+				|| !g_isi_sb_iter_get_alpha_tag(&iter, &tag,
 						taglen * 2, 6))
 				goto error;
 
@@ -433,17 +432,17 @@ static bool available_resp_cb(GIsiClient *client, const void *restrict data,
 			op->name[OFONO_MAX_OPERATOR_NAME_LENGTH] = '\0';
 			g_free(tag);
 			break;
-		}
-
-		case NET_DETAILED_NETWORK_INFO: {
-			struct ofono_network_operator *op;
 
+		case NET_DETAILED_NETWORK_INFO:
 			op = list + detail++;
+
 			if (!g_isi_sb_iter_get_oper_code(&iter, op->mcc,
-								op->mnc, 2))
+								op->mnc, 2)
+				|| !g_isi_sb_iter_get_byte(&iter, &umts, 7))
 				goto error;
+
+			op->tech = umts ? 2 : 3;
 			break;
-		}
 
 		default:
 			DBG("Skipping sub-block: %s (%zu bytes)",
@@ -451,7 +450,6 @@ static bool available_resp_cb(GIsiClient *client, const void *restrict data,
 				g_isi_sb_iter_get_len(&iter));
 			break;
 		}
-		g_isi_sb_iter_next(&iter);
 	}
 
 	if (common == detail && detail == total) {
@@ -464,7 +462,7 @@ error:
 
 out:
 	g_free(cbd);
-	return true;
+	return TRUE;
 }
 
 static void isi_list_operators(struct ofono_netreg *netreg,
@@ -484,7 +482,7 @@ static void isi_list_operators(struct ofono_netreg *netreg,
 		0x00
 	};
 
-	if (!cbd)
+	if (!cbd || !net)
 		goto error;
 
 	if (g_isi_request_make(net->client, msg, sizeof(msg),
@@ -497,8 +495,9 @@ error:
 	g_free(cbd);
 }
 
-static bool set_auto_resp_cb(GIsiClient *client, const void *restrict data,
-				size_t len, uint16_t object, void *opaque)
+static gboolean set_auto_resp_cb(GIsiClient *client,
+					const void *restrict data, size_t len,
+					uint16_t object, void *opaque)
 {
 	const unsigned char *msg = data;
 	struct isi_cb_data *cbd = opaque;
@@ -511,7 +510,7 @@ static bool set_auto_resp_cb(GIsiClient *client, const void *restrict data,
 	}
 
 	if (!msg || len < 3 || msg[0] != NET_SET_RESP)
-		goto error;
+		return FALSE;
 
 	if (msg[1] != NET_CAUSE_OK) {
 		DBG("Request failed: %s", net_isi_cause_name(msg[1]));
@@ -527,7 +526,7 @@ error:
 
 out:
 	g_free(cbd);
-	return true;
+	return TRUE;
 }
 
 static void isi_register_auto(struct ofono_netreg *netreg,
@@ -549,7 +548,7 @@ static void isi_register_auto(struct ofono_netreg *netreg,
 		0x00  /* Index not used */
 	};
 
-	if (!cbd)
+	if (!cbd || !net)
 		goto error;
 
 	if (g_isi_request_make(net->client, msg, sizeof(msg),
@@ -562,8 +561,9 @@ error:
 	g_free(cbd);
 }
 
-static bool set_manual_resp_cb(GIsiClient *client, const void *restrict data,
-				size_t len, uint16_t object, void *opaque)
+static gboolean set_manual_resp_cb(GIsiClient *client,
+					const void *restrict data, size_t len,
+					uint16_t object, void *opaque)
 {
 	const unsigned char *msg = data;
 	struct isi_cb_data *cbd = opaque;
@@ -571,13 +571,13 @@ static bool set_manual_resp_cb(GIsiClient *client, const void *restrict data,
 	struct netreg_data *nd = ofono_netreg_get_data(netreg);
 	ofono_netreg_register_cb_t cb = cbd->cb;
 
-	if(!msg) {
+	if (!msg) {
 		DBG("ISI client error: %d", g_isi_client_error(client));
 		goto error;
 	}
 
 	if (len < 3 || msg[0] != NET_SET_RESP)
-		goto error;
+		return FALSE;
 
 	if (msg[1] != NET_CAUSE_OK) {
 		DBG("Request failed: %s", net_isi_cause_name(msg[1]));
@@ -593,7 +593,7 @@ error:
 
 out:
 	g_free(cbd);
-	return true;
+	return TRUE;
 }
 
 static void isi_register_manual(struct ofono_netreg *netreg,
@@ -621,7 +621,7 @@ static void isi_register_manual(struct ofono_netreg *netreg,
 		0x00, 0x00  /* Filler */
 	};
 
-	if (!cbd)
+	if (!cbd || !nd)
 		goto error;
 
 	if (g_isi_request_make(nd->client, msg, sizeof(msg),
@@ -642,8 +642,9 @@ static void isi_deregister(struct ofono_netreg *netreg,
 	CALLBACK_WITH_FAILURE(cb, data);
 }
 
-static void rat_ind_cb(GIsiClient *client, const void *restrict data,
-			size_t len, uint16_t object, void *opaque)
+static void rat_ind_cb(GIsiClient *client,
+			const void *restrict data, size_t len,
+			uint16_t object, void *opaque)
 {
 	const unsigned char *msg = data;
 	struct ofono_netreg *netreg = opaque;
@@ -663,16 +664,16 @@ static void rat_ind_cb(GIsiClient *client, const void *restrict data,
 		case NET_RAT_INFO: {
 			guint8 info = 0;
 
-			if (!g_isi_sb_iter_get_byte(&iter, &nd->rat, 2))
+			if (!g_isi_sb_iter_get_byte(&iter, &nd->rat, 2)
+				|| !g_isi_sb_iter_get_byte(&iter, &info, 3))
 				return;
 
-			if (!g_isi_sb_iter_get_byte(&iter, &info, 3))
-				return;
+			if (info) {
 
-			if (info)
 				if (!g_isi_sb_iter_get_byte(&iter,
 							&nd->gsm_compact, 4))
 					return;
+			}
 			break;
 		}
 
@@ -686,8 +687,9 @@ static void rat_ind_cb(GIsiClient *client, const void *restrict data,
 	}
 }
 
-static bool rat_resp_cb(GIsiClient *client, const void *restrict data,
-			size_t len, uint16_t object, void *opaque)
+static gboolean rat_resp_cb(GIsiClient *client,
+				const void *restrict data, size_t len,
+				uint16_t object, void *opaque)
 {
 	const unsigned char *msg = data;
 	struct ofono_netreg *netreg = opaque;
@@ -697,15 +699,15 @@ static bool rat_resp_cb(GIsiClient *client, const void *restrict data,
 
 	if (!msg) {
 		DBG("ISI client error: %d", g_isi_client_error(client));
-		return true;
+		return TRUE;
 	}
 
 	if (len < 3 || msg[0] != NET_RAT_RESP)
-		return false;
+		return FALSE;
 
 	if (msg[1] != NET_CAUSE_OK) {
 		DBG("Request failed: %s", net_isi_cause_name(msg[1]));
-		return true;
+		return TRUE;
 	}
 
 	g_isi_sb_iter_init(&iter, msg, len, 3);
@@ -722,7 +724,7 @@ static bool rat_resp_cb(GIsiClient *client, const void *restrict data,
 				|| !info
 				|| !g_isi_sb_iter_get_byte(&iter,
 							&nd->gsm_compact, 4))
-				return true;
+				return TRUE;
 
 			break;
 		}
@@ -735,11 +737,12 @@ static bool rat_resp_cb(GIsiClient *client, const void *restrict data,
 		}
 		g_isi_sb_iter_next(&iter);
 	}
-	return true;
+	return TRUE;
 }
 
-static void rssi_ind_cb(GIsiClient *client, const void *restrict data,
-			size_t len, uint16_t object, void *opaque)
+static void rssi_ind_cb(GIsiClient *client,
+			const void *restrict data, size_t len,
+			uint16_t object, void *opaque)
 {
 	const unsigned char *msg = data;
 	struct ofono_netreg *netreg = opaque;
@@ -750,8 +753,48 @@ static void rssi_ind_cb(GIsiClient *client, const void *restrict data,
 	ofono_netreg_strength_notify(netreg, msg[1]);
 }
 
-static bool rssi_resp_cb(GIsiClient *client, const void *restrict data,
-				size_t len, uint16_t object, void *opaque)
+static void time_ind_cb(GIsiClient *client,
+			const void *restrict data, size_t len,
+			uint16_t object, void *opaque)
+{
+	const unsigned char *msg = data;
+	const unsigned char *nitz = msg + 3;
+	struct ofono_netreg *netreg = opaque;
+
+	struct ofono_network_time info;
+
+	if (!msg || len < 13 || msg[0] != NET_TIME_IND
+		|| nitz[0] != NET_TIME_INFO)
+		return;
+
+	nitz += 2;
+
+	/* Value is years since last turn of century */
+	info.year = nitz[0] != NET_INVALID_TIME ? nitz[0] : -1;
+	info.year += 2000;
+
+	info.mon = nitz[1] != NET_INVALID_TIME ? nitz[1] : -1;
+	info.mday = nitz[2] != NET_INVALID_TIME ? nitz[2] : -1;
+	info.hour = nitz[3] != NET_INVALID_TIME ? nitz[3] : -1;
+	info.min = nitz[4] != NET_INVALID_TIME ? nitz[4] : -1;
+	info.sec = nitz[5] != NET_INVALID_TIME ? nitz[5] : -1;
+
+	/* Most significant bit set indicates negative offset. The
+	 * second most significant bit is 'reserved'. The value is the
+	 * offset from UTCin a count of 15min intervals, possibly
+	 * including the current DST adjustment. */
+	info.utcoff = (nitz[6] & 0x3F) * 15 * 60;
+	if (nitz[6] & 0x80)
+		info.utcoff *= -1;
+
+	info.dst = nitz[7] != NET_INVALID_TIME ? nitz[7] : -1;
+
+	ofono_netreg_time_notify(netreg, &info);
+}
+
+static gboolean rssi_resp_cb(GIsiClient *client,
+				const void *restrict data, size_t len,
+				uint16_t object, void *opaque)
 {
 	const unsigned char *msg = data;
 	struct isi_cb_data *cbd = opaque;
@@ -766,10 +809,11 @@ static bool rssi_resp_cb(GIsiClient *client, const void *restrict data,
 	}
 
 	if (len < 3 || msg[0] != NET_RSSI_GET_RESP)
-		return false;
+		return FALSE;
 
 	if (msg[1] != NET_CAUSE_OK) {
-		DBG("Request failed: %s", net_isi_cause_name(msg[1]));
+		DBG("Request failed: %s (0x%02X)",
+			net_isi_cause_name(msg[1]), msg[1]);
 		goto error;
 	}
 
@@ -806,7 +850,7 @@ error:
 
 out:
 	g_free(cbd);
-	return true;
+	return TRUE;
 }
 
 static void isi_strength(struct ofono_netreg *netreg,
@@ -822,7 +866,7 @@ static void isi_strength(struct ofono_netreg *netreg,
 		NET_CURRENT_CELL_RSSI
 	};
 
-	if (!cbd)
+	if (!cbd || !nd)
 		goto error;
 
 	if (g_isi_request_make(nd->client, msg, sizeof(msg),
@@ -854,6 +898,7 @@ static gboolean isi_netreg_register(gpointer user)
 	g_isi_subscribe(nd->client, NET_REG_STATUS_IND, reg_status_ind_cb,
 			netreg);
 	g_isi_subscribe(nd->client, NET_RAT_IND, rat_ind_cb, netreg);
+	g_isi_subscribe(nd->client, NET_TIME_IND, time_ind_cb, netreg);
 
 	/* Bootstrap current RAT setting */
 	if (!g_isi_request_make(nd->client, rat, sizeof(rat),
@@ -866,13 +911,13 @@ static gboolean isi_netreg_register(gpointer user)
 	return FALSE;
 }
 
-static void reachable_cb(GIsiClient *client, bool alive, uint16_t object,
+static void reachable_cb(GIsiClient *client, gboolean alive, uint16_t object,
 				void *opaque)
 {
 	struct ofono_netreg *netreg = opaque;
 
 	if (!alive) {
-		DBG("Unable to bootsrap netreg driver");
+		DBG("Unable to bootstrap netreg driver");
 		return;
 	}
 
@@ -909,10 +954,12 @@ static void isi_netreg_remove(struct ofono_netreg *net)
 {
 	struct netreg_data *data = ofono_netreg_get_data(net);
 
-	if (data) {
-		g_isi_client_destroy(data->client);
-		g_free(data);
-	}
+	if (!data)
+		return;
+
+	ofono_netreg_set_data(net, NULL);
+	g_isi_client_destroy(data->client);
+	g_free(data);
 }
 
 static struct ofono_netreg_driver driver = {
