@@ -2,7 +2,7 @@
  *
  *  oFono - Open Source Telephony
  *
- *  Copyright (C) 2008-2009  Intel Corporation. All rights reserved.
+ *  Copyright (C) 2008-2010  Intel Corporation. All rights reserved.
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License version 2 as
@@ -27,37 +27,44 @@
 #include <stdio.h>
 #include <stdlib.h>
 #include <time.h>
+#include <errno.h>
 
 #include <glib.h>
 #include <gdbus.h>
 
 #include "ofono.h"
 
-#include "driver.h"
 #include "common.h"
-#include "cssn.h"
-#include "ussd.h"
-
-#define CALL_BARRING_INTERFACE "org.ofono.CallBarring"
 
 #define CALL_BARRING_FLAG_CACHED 0x1
+#define NUM_OF_BARRINGS 5
 
-static void cb_ss_query_next_lock(struct ofono_modem *modem);
-static void get_query_next_lock(struct ofono_modem *modem);
-static void set_query_next_lock(struct ofono_modem *modem);
+static GSList *g_drivers = NULL;
 
-struct call_barring_data {
-	struct ofono_call_barring_ops *ops;
+static void cb_ss_query_next_lock(struct ofono_call_barring *cb);
+static void get_query_next_lock(struct ofono_call_barring *cb);
+static void set_query_next_lock(struct ofono_call_barring *cb);
+
+struct ofono_call_barring {
 	int flags;
 	DBusMessage *pending;
-	int *cur_locks;
-	int *new_locks;
+	int cur_locks[NUM_OF_BARRINGS];
+	int new_locks[NUM_OF_BARRINGS];
 	int query_start;
 	int query_end;
 	int query_next;
 	int ss_req_type;
 	int ss_req_cls;
 	int ss_req_lock;
+	struct ofono_ssn *ssn;
+	struct ofono_ussd *ussd;
+	unsigned int incoming_bar_watch;
+	unsigned int outgoing_bar_watch;
+	unsigned int ssn_watch;
+	unsigned int ussd_watch;
+	const struct ofono_call_barring_driver *driver;
+	void *driver_data;
+	struct ofono_atom *atom;
 };
 
 struct call_barring_lock {
@@ -88,11 +95,12 @@ static struct call_barring_lock cb_locks[] = {
 #define CB_ALL_OUTGOING 6
 #define CB_ALL_INCOMING 7
 
-static inline void emit_barring_changed(struct ofono_modem *modem, int start,
-					int end, const char *type, int cls)
+static inline void emit_barring_changed(struct ofono_call_barring *cb,
+					int start, int end,
+					const char *type, int cls)
 {
-	struct call_barring_data *cb = modem->call_barring;
 	DBusConnection *conn = ofono_dbus_get_connection();
+	const char *path = __ofono_atom_get_path(cb->atom);
 	char property_name[64];
 	const char *value;
 	int i;
@@ -117,15 +125,14 @@ static inline void emit_barring_changed(struct ofono_modem *modem, int start,
 	snprintf(property_name, sizeof(property_name), "%s%s",
 			bearer_class_to_string(cls), type);
 
-	ofono_dbus_signal_property_changed(conn, modem->path,
-						CALL_BARRING_INTERFACE,
+	ofono_dbus_signal_property_changed(conn, path,
+						OFONO_CALL_BARRING_INTERFACE,
 						property_name, DBUS_TYPE_STRING,
 						&value);
 }
 
-static void update_barrings(struct ofono_modem *modem, int mask)
+static void update_barrings(struct ofono_call_barring *cb, int mask)
 {
-	struct call_barring_data *cb = modem->call_barring;
 	int cls;
 	int i;
 
@@ -134,19 +141,17 @@ static void update_barrings(struct ofono_modem *modem, int mask)
 		if ((cls & mask) == 0)
 			continue;
 
-		emit_barring_changed(modem, cb->query_start, CB_OUTGOING_END,
+		emit_barring_changed(cb, cb->query_start, CB_OUTGOING_END,
 					"Outgoing", cls);
-		emit_barring_changed(modem, CB_INCOMING_START, cb->query_end,
+		emit_barring_changed(cb, CB_INCOMING_START, cb->query_end,
 					"Incoming", cls);
 	}
 
-	for (i = cb->query_start; i <= cb->query_end; i++) {
+	for (i = cb->query_start; i <= cb->query_end; i++)
 		cb->cur_locks[i] = cb->new_locks[i];
-		cb->new_locks[i] = 0;
-	}
 }
 
-static void cb_ss_property_append(struct call_barring_data *cb,
+static void cb_ss_property_append(struct ofono_call_barring *cb,
 					DBusMessageIter *dict, int lock,
 					int mask)
 {
@@ -158,7 +163,7 @@ static void cb_ss_property_append(struct call_barring_data *cb,
 		if (!(mask & i))
 			continue;
 
-		strvalue = (cb->cur_locks[lock] & i) ? "enabled" : "disabled";
+		strvalue = (cb->new_locks[lock] & i) ? "enabled" : "disabled";
 
 		snprintf(property_name, sizeof(property_name), "%s%s",
 				bearer_class_to_string(i),
@@ -169,7 +174,7 @@ static void cb_ss_property_append(struct call_barring_data *cb,
 	}
 }
 
-static void cb_set_query_bounds(struct call_barring_data *cb,
+static void cb_set_query_bounds(struct ofono_call_barring *cb,
 				const char *fac, gboolean fac_only)
 {
 	int i;
@@ -222,9 +227,8 @@ incoming:
 	return;
 }
 
-static void generate_ss_query_reply(struct ofono_modem *modem)
+static void generate_ss_query_reply(struct ofono_call_barring *cb)
 {
-	struct call_barring_data *cb = modem->call_barring;
 	const char *context = "CallBarring";
 	const char *sig = "(ssa{sv})";
 	const char *ss_type = ss_control_type_to_string(cb->ss_req_type);
@@ -279,10 +283,9 @@ static void generate_ss_query_reply(struct ofono_modem *modem)
 }
 
 static void cb_ss_query_next_lock_callback(const struct ofono_error *error,
-				int status, void *data)
+					int status, void *data)
 {
-	struct ofono_modem *modem = data;
-	struct call_barring_data *cb = modem->call_barring;
+	struct ofono_call_barring *cb = data;
 
 	if (error->type != OFONO_ERROR_TYPE_NO_ERROR) {
 		if (cb->ss_req_type != SS_CONTROL_TYPE_QUERY)
@@ -300,40 +303,38 @@ static void cb_ss_query_next_lock_callback(const struct ofono_error *error,
 
 	if (cb->query_next < cb->query_end) {
 		cb->query_next += 1;
-		cb_ss_query_next_lock(modem);
+		cb_ss_query_next_lock(cb);
 		return;
 	}
 
-	generate_ss_query_reply(modem);
-	update_barrings(modem, BEARER_CLASS_VOICE);
+	generate_ss_query_reply(cb);
+	update_barrings(cb, BEARER_CLASS_VOICE);
 }
 
-static void cb_ss_query_next_lock(struct ofono_modem *modem)
+static void cb_ss_query_next_lock(struct ofono_call_barring *cb)
 {
-	struct call_barring_data *cb = modem->call_barring;
 	int cls;
 
 	cls = cb->ss_req_cls | BEARER_CLASS_DEFAULT;
 
-	cb->ops->query(modem, cb_locks[cb->query_next].fac, cls,
-			cb_ss_query_next_lock_callback, modem);
+	cb->driver->query(cb, cb_locks[cb->query_next].fac, cls,
+			cb_ss_query_next_lock_callback, cb);
 }
 
 static void cb_ss_set_lock_callback(const struct ofono_error *error,
 		void *data)
 {
-	struct ofono_modem *modem = data;
-	struct call_barring_data *cb = modem->call_barring;
+	struct ofono_call_barring *cb = data;
 
 	if (error->type != OFONO_ERROR_TYPE_NO_ERROR) {
-		ofono_debug("Enabling/disabling Call Barring via SS failed");
+		DBG("Enabling/disabling Call Barring via SS failed");
 		__ofono_dbus_pending_reply(&cb->pending,
 					__ofono_error_failed(cb->pending));
 		return;
 	}
 
 	/* Assume we have query always */
-	cb_ss_query_next_lock(modem);
+	cb_ss_query_next_lock(cb);
 }
 
 static const char *cb_ss_service_to_fac(const char *svc)
@@ -358,13 +359,12 @@ static const char *cb_ss_service_to_fac(const char *svc)
 	return NULL;
 }
 
-static gboolean cb_ss_control(struct ofono_modem *modem,
-				enum ss_control_type type, const char *sc,
+static gboolean cb_ss_control(int type, const char *sc,
 				const char *sia, const char *sib,
 				const char *sic, const char *dn,
-				DBusMessage *msg)
+				DBusMessage *msg, void *data)
 {
-	struct call_barring_data *cb = modem->call_barring;
+	struct ofono_call_barring *cb = data;
 	DBusConnection *conn = ofono_dbus_get_connection();
 	int cls = BEARER_CLASS_DEFAULT;
 	const char *fac;
@@ -372,17 +372,17 @@ static gboolean cb_ss_control(struct ofono_modem *modem,
 	void *operation = NULL;
 	int i;
 
-	if (cb->pending) {
+	if (__ofono_call_barring_is_busy(cb)) {
 		reply = __ofono_error_busy(msg);
 		g_dbus_send_message(conn, reply);
 
 		return TRUE;
 	}
 
-	ofono_debug("Received call barring ss control request");
+	DBG("Received call barring ss control request");
 
-	ofono_debug("type: %d, sc: %s, sia: %s, sib: %s, sic: %s, dn: %s",
-			type, sc, sia, sib, sic, dn);
+	DBG("type: %d, sc: %s, sia: %s, sib: %s, sic: %s, dn: %s",
+		type, sc, sia, sib, sic, dn);
 
 	fac = cb_ss_service_to_fac(sc);
 	if (!fac)
@@ -402,7 +402,7 @@ static gboolean cb_ss_control(struct ofono_modem *modem,
 	if (strlen(dn) > 0)
 		goto bad_format;
 
-	if (!is_valid_pin(sia))
+	if (type != SS_CONTROL_TYPE_QUERY && !is_valid_pin(sia, PIN_TYPE_NET))
 		goto bad_format;
 
 	switch (type) {
@@ -410,10 +410,12 @@ static gboolean cb_ss_control(struct ofono_modem *modem,
 	case SS_CONTROL_TYPE_DEACTIVATION:
 	case SS_CONTROL_TYPE_REGISTRATION:
 	case SS_CONTROL_TYPE_ERASURE:
-		operation = cb->ops->set;
+		operation = cb->driver->set;
 		break;
 	case SS_CONTROL_TYPE_QUERY:
-		operation = cb->ops->query;
+		operation = cb->driver->query;
+		break;
+	default:
 		break;
 	}
 
@@ -424,7 +426,8 @@ static gboolean cb_ss_control(struct ofono_modem *modem,
 		return TRUE;
 	}
 
-	/* According to 27.007, AG, AC and AB only work with mode = 0
+	/*
+	 * According to 27.007, AG, AC and AB only work with mode = 0
 	 * We support query by querying all relevant types, since we must
 	 * do this for the deactivation case anyway
 	 */
@@ -454,19 +457,19 @@ static gboolean cb_ss_control(struct ofono_modem *modem,
 	switch (type) {
 	case SS_CONTROL_TYPE_ACTIVATION:
 	case SS_CONTROL_TYPE_REGISTRATION:
-		cb->ss_req_type = SS_CONTROL_TYPE_REGISTRATION;
-		cb->ops->set(modem, fac, 1, sia, cls,
-				cb_ss_set_lock_callback, modem);
+		cb->ss_req_type = SS_CONTROL_TYPE_ACTIVATION;
+		cb->driver->set(cb, fac, 1, sia, cls,
+				cb_ss_set_lock_callback, cb);
 		break;
 	case SS_CONTROL_TYPE_ERASURE:
 	case SS_CONTROL_TYPE_DEACTIVATION:
-		cb->ss_req_type = SS_CONTROL_TYPE_ERASURE;
-		cb->ops->set(modem, fac, 0, sia, cls,
-				cb_ss_set_lock_callback, modem);
+		cb->ss_req_type = SS_CONTROL_TYPE_DEACTIVATION;
+		cb->driver->set(cb, fac, 0, sia, cls,
+				cb_ss_set_lock_callback, cb);
 		break;
 	case SS_CONTROL_TYPE_QUERY:
 		cb->ss_req_type = SS_CONTROL_TYPE_QUERY;
-		cb_ss_query_next_lock(modem);
+		cb_ss_query_next_lock(cb);
 		break;
 	}
 
@@ -480,39 +483,38 @@ bad_format:
 
 static void cb_set_passwd_callback(const struct ofono_error *error, void *data)
 {
-	struct ofono_modem *modem = data;
-	struct call_barring_data *cb = modem->call_barring;
+	struct ofono_call_barring *cb = data;
 	DBusMessage *reply;
 
 	if (error->type == OFONO_ERROR_TYPE_NO_ERROR)
 		reply = dbus_message_new_method_return(cb->pending);
 	else {
 		reply = __ofono_error_failed(cb->pending);
-		ofono_debug("Changing Call Barring password via SS failed");
+		DBG("Changing Call Barring password via SS failed");
 	}
 
 	__ofono_dbus_pending_reply(&cb->pending, reply);
 }
 
-static gboolean cb_ss_passwd(struct ofono_modem *modem, const char *sc,
+static gboolean cb_ss_passwd(const char *sc,
 				const char *old, const char *new,
-				DBusMessage *msg)
+				DBusMessage *msg, void *data)
 {
-	struct call_barring_data *cb = modem->call_barring;
+	struct ofono_call_barring *cb = data;
 	DBusConnection *conn = ofono_dbus_get_connection();
 	DBusMessage *reply;
 	const char *fac;
 
-	if (cb->pending) {
+	if (__ofono_call_barring_is_busy(cb)) {
 		reply = __ofono_error_busy(msg);
 		g_dbus_send_message(conn, reply);
 
 		return TRUE;
 	}
 
-	ofono_debug("Received call barring ss password change request");
+	DBG("Received call barring ss password change request");
 
-	ofono_debug("sc: %s", sc);
+	DBG("sc: %s", sc);
 
 	if (!strcmp(sc, ""))
 		fac = "AB";
@@ -522,12 +524,11 @@ static gboolean cb_ss_passwd(struct ofono_modem *modem, const char *sc,
 	if (!fac)
 		return FALSE;
 
-	if (!is_valid_pin(old) || !is_valid_pin(new))
+	if (!is_valid_pin(old, PIN_TYPE_NET) || !is_valid_pin(new, PIN_TYPE_NET))
 		goto bad_format;
 
 	cb->pending = dbus_message_ref(msg);
-	cb->ops->set_passwd(modem, fac, old, new,
-			cb_set_passwd_callback, modem);
+	cb->driver->set_passwd(cb, fac, old, new, cb_set_passwd_callback, cb);
 
 	return TRUE;
 bad_format:
@@ -536,74 +537,56 @@ bad_format:
 	return TRUE;
 }
 
-static void cb_register_ss_controls(struct ofono_modem *modem)
+static void cb_register_ss_controls(struct ofono_call_barring *cb)
 {
-	ss_control_register(modem, "33", cb_ss_control);
-	ss_control_register(modem, "331", cb_ss_control);
-	ss_control_register(modem, "332", cb_ss_control);
-	ss_control_register(modem, "35", cb_ss_control);
-	ss_control_register(modem, "351", cb_ss_control);
-	ss_control_register(modem, "330", cb_ss_control);
-	ss_control_register(modem, "333", cb_ss_control);
-	ss_control_register(modem, "353", cb_ss_control);
-	ss_passwd_register(modem, "", cb_ss_passwd);
-	ss_passwd_register(modem, "33", cb_ss_passwd);
-	ss_passwd_register(modem, "331", cb_ss_passwd);
-	ss_passwd_register(modem, "332", cb_ss_passwd);
-	ss_passwd_register(modem, "35", cb_ss_passwd);
-	ss_passwd_register(modem, "351", cb_ss_passwd);
-	ss_passwd_register(modem, "330", cb_ss_passwd);
-	ss_passwd_register(modem, "333", cb_ss_passwd);
-	ss_passwd_register(modem, "353", cb_ss_passwd);
+	__ofono_ussd_ssc_register(cb->ussd, "33", cb_ss_control, cb, NULL);
+	__ofono_ussd_ssc_register(cb->ussd, "331", cb_ss_control, cb, NULL);
+	__ofono_ussd_ssc_register(cb->ussd, "332", cb_ss_control, cb, NULL);
+	__ofono_ussd_ssc_register(cb->ussd, "35", cb_ss_control, cb, NULL);
+	__ofono_ussd_ssc_register(cb->ussd, "351", cb_ss_control, cb, NULL);
+	__ofono_ussd_ssc_register(cb->ussd, "330", cb_ss_control, cb, NULL);
+	__ofono_ussd_ssc_register(cb->ussd, "333", cb_ss_control, cb, NULL);
+	__ofono_ussd_ssc_register(cb->ussd, "353", cb_ss_control, cb, NULL);
+
+	__ofono_ussd_passwd_register(cb->ussd, "", cb_ss_passwd, cb, NULL);
+	__ofono_ussd_passwd_register(cb->ussd, "33", cb_ss_passwd, cb, NULL);
+	__ofono_ussd_passwd_register(cb->ussd, "331", cb_ss_passwd, cb, NULL);
+	__ofono_ussd_passwd_register(cb->ussd, "332", cb_ss_passwd, cb, NULL);
+	__ofono_ussd_passwd_register(cb->ussd, "35", cb_ss_passwd, cb, NULL);
+	__ofono_ussd_passwd_register(cb->ussd, "351", cb_ss_passwd, cb, NULL);
+	__ofono_ussd_passwd_register(cb->ussd, "330", cb_ss_passwd, cb, NULL);
+	__ofono_ussd_passwd_register(cb->ussd, "333", cb_ss_passwd, cb, NULL);
+	__ofono_ussd_passwd_register(cb->ussd, "353", cb_ss_passwd, cb, NULL);
 }
 
-static void cb_unregister_ss_controls(struct ofono_modem *modem)
+static void cb_unregister_ss_controls(struct ofono_call_barring *cb)
 {
-	ss_control_unregister(modem, "33", cb_ss_control);
-	ss_control_unregister(modem, "331", cb_ss_control);
-	ss_control_unregister(modem, "332", cb_ss_control);
-	ss_control_unregister(modem, "35", cb_ss_control);
-	ss_control_unregister(modem, "351", cb_ss_control);
-	ss_control_unregister(modem, "330", cb_ss_control);
-	ss_control_unregister(modem, "333", cb_ss_control);
-	ss_control_unregister(modem, "353", cb_ss_control);
-	ss_passwd_unregister(modem, "", cb_ss_passwd);
-	ss_passwd_unregister(modem, "33", cb_ss_passwd);
-	ss_passwd_unregister(modem, "331", cb_ss_passwd);
-	ss_passwd_unregister(modem, "332", cb_ss_passwd);
-	ss_passwd_unregister(modem, "35", cb_ss_passwd);
-	ss_passwd_unregister(modem, "351", cb_ss_passwd);
-	ss_passwd_unregister(modem, "330", cb_ss_passwd);
-	ss_passwd_unregister(modem, "333", cb_ss_passwd);
-	ss_passwd_unregister(modem, "353", cb_ss_passwd);
+	__ofono_ussd_ssc_unregister(cb->ussd, "33");
+	__ofono_ussd_ssc_unregister(cb->ussd, "331");
+	__ofono_ussd_ssc_unregister(cb->ussd, "332");
+	__ofono_ussd_ssc_unregister(cb->ussd, "35");
+	__ofono_ussd_ssc_unregister(cb->ussd, "351");
+	__ofono_ussd_ssc_unregister(cb->ussd, "330");
+	__ofono_ussd_ssc_unregister(cb->ussd, "333");
+	__ofono_ussd_ssc_unregister(cb->ussd, "353");
+
+	__ofono_ussd_passwd_unregister(cb->ussd, "");
+	__ofono_ussd_passwd_unregister(cb->ussd, "33");
+	__ofono_ussd_passwd_unregister(cb->ussd, "331");
+	__ofono_ussd_passwd_unregister(cb->ussd, "332");
+	__ofono_ussd_passwd_unregister(cb->ussd, "35");
+	__ofono_ussd_passwd_unregister(cb->ussd, "351");
+	__ofono_ussd_passwd_unregister(cb->ussd, "330");
+	__ofono_ussd_passwd_unregister(cb->ussd, "333");
+	__ofono_ussd_passwd_unregister(cb->ussd, "353");
 }
 
-static struct call_barring_data *call_barring_create(void)
+gboolean __ofono_call_barring_is_busy(struct ofono_call_barring *cb)
 {
-	int lcount;
-	struct call_barring_data *cb = g_new0(struct call_barring_data, 1);
-
-	lcount = CB_ALL_END - CB_ALL_START + 1;
-
-	cb->cur_locks = g_new0(int, lcount);
-	cb->new_locks = g_new0(int, lcount);
-
-	return cb;
+	return cb->pending ? TRUE : FALSE;
 }
 
-static void call_barring_destroy(gpointer userdata)
-{
-	struct ofono_modem *modem = userdata;
-	struct call_barring_data *cb = modem->call_barring;
-
-	g_free(cb->cur_locks);
-	g_free(cb->new_locks);
-	g_free(cb);
-
-	modem->call_barring = NULL;
-}
-
-static inline void cb_append_property(struct call_barring_data *cb,
+static inline void cb_append_property(struct ofono_call_barring *cb,
 					DBusMessageIter *dict, int start,
 					int end, int cls, const char *property)
 {
@@ -612,7 +595,7 @@ static inline void cb_append_property(struct call_barring_data *cb,
 	int i;
 
 	for (i = start; i <= end; i++)
-		if (cb->cur_locks[i] & cls)
+		if (cb->new_locks[i] & cls)
 			break;
 
 	if (i <= end)
@@ -625,9 +608,8 @@ static inline void cb_append_property(struct call_barring_data *cb,
 				&value);
 }
 
-static void cb_get_properties_reply(struct ofono_modem *modem, int mask)
+static void cb_get_properties_reply(struct ofono_call_barring *cb, int mask)
 {
-	struct call_barring_data *cb = modem->call_barring;
 	DBusMessage *reply;
 	DBusMessageIter iter, dict;
 	int j;
@@ -663,8 +645,7 @@ static void cb_get_properties_reply(struct ofono_modem *modem, int mask)
 static void get_query_lock_callback(const struct ofono_error *error,
 					int status, void *data)
 {
-	struct ofono_modem *modem = data;
-	struct call_barring_data *cb = modem->call_barring;
+	struct ofono_call_barring *cb = data;
 
 	if (error->type == OFONO_ERROR_TYPE_NO_ERROR) {
 		cb->new_locks[cb->query_next] = status;
@@ -675,41 +656,38 @@ static void get_query_lock_callback(const struct ofono_error *error,
 
 	if (cb->query_next < CB_ALL_END) {
 		cb->query_next = cb->query_next + 1;
-		get_query_next_lock(modem);
+		get_query_next_lock(cb);
 		return;
 	}
 
-	cb_get_properties_reply(modem, BEARER_CLASS_VOICE);
-	update_barrings(modem, BEARER_CLASS_VOICE);
+	cb_get_properties_reply(cb, BEARER_CLASS_VOICE);
+	update_barrings(cb, BEARER_CLASS_VOICE);
 }
 
-static void get_query_next_lock(struct ofono_modem *modem)
+static void get_query_next_lock(struct ofono_call_barring *cb)
 {
-	struct call_barring_data *cb = modem->call_barring;
-
-	cb->ops->query(modem, cb_locks[cb->query_next].fac,
-			BEARER_CLASS_DEFAULT, get_query_lock_callback, modem);
+	cb->driver->query(cb, cb_locks[cb->query_next].fac,
+			BEARER_CLASS_DEFAULT, get_query_lock_callback, cb);
 }
 
 static DBusMessage *cb_get_properties(DBusConnection *conn, DBusMessage *msg,
 					void *data)
 {
-	struct ofono_modem *modem = data;
-	struct call_barring_data *cb = modem->call_barring;
+	struct ofono_call_barring *cb = data;
 
-	if (cb->pending)
+	if (__ofono_call_barring_is_busy(cb) || __ofono_ussd_is_busy(cb->ussd))
 		return __ofono_error_busy(msg);
 
-	if (!cb->ops->query)
+	if (!cb->driver->query)
 		return __ofono_error_not_implemented(msg);
 
 	cb->pending = dbus_message_ref(msg);
 
 	if (cb->flags & CALL_BARRING_FLAG_CACHED)
-		cb_get_properties_reply(modem, BEARER_CLASS_VOICE);
+		cb_get_properties_reply(cb, BEARER_CLASS_VOICE);
 	else {
 		cb->query_next = CB_ALL_START;
-		get_query_next_lock(modem);
+		get_query_next_lock(cb);
 	}
 
 	return NULL;
@@ -718,8 +696,7 @@ static DBusMessage *cb_get_properties(DBusConnection *conn, DBusMessage *msg,
 static void set_query_lock_callback(const struct ofono_error *error,
 				int status, void *data)
 {
-	struct ofono_modem *modem = data;
-	struct call_barring_data *cb = modem->call_barring;
+	struct ofono_call_barring *cb = data;
 
 	if (error->type != OFONO_ERROR_TYPE_NO_ERROR) {
 		ofono_error("Disabling all barring successful, "
@@ -736,36 +713,34 @@ static void set_query_lock_callback(const struct ofono_error *error,
 
 	if (cb->query_next < cb->query_end) {
 		cb->query_next += 1;
-		set_query_next_lock(modem);
+		set_query_next_lock(cb);
 		return;
 	}
 
 	__ofono_dbus_pending_reply(&cb->pending,
 				dbus_message_new_method_return(cb->pending));
-	update_barrings(modem, BEARER_CLASS_VOICE);
+	update_barrings(cb, BEARER_CLASS_VOICE);
 }
 
-static void set_query_next_lock(struct ofono_modem *modem)
+static void set_query_next_lock(struct ofono_call_barring *cb)
 {
-	struct call_barring_data *cb = modem->call_barring;
-
-	cb->ops->query(modem, cb_locks[cb->query_next].fac,
-			BEARER_CLASS_DEFAULT, set_query_lock_callback, modem);
+	cb->driver->query(cb, cb_locks[cb->query_next].fac,
+			BEARER_CLASS_DEFAULT, set_query_lock_callback, cb);
 }
 
 static void set_lock_callback(const struct ofono_error *error, void *data)
 {
-	struct ofono_modem *modem = data;
-	struct call_barring_data *cb = modem->call_barring;
+	struct ofono_call_barring *cb = data;
 
 	if (error->type != OFONO_ERROR_TYPE_NO_ERROR) {
-		ofono_debug("Enabling/disabling a lock failed");
+		DBG("Enabling/disabling a lock failed");
 		__ofono_dbus_pending_reply(&cb->pending,
 					__ofono_error_failed(cb->pending));
 		return;
 	}
 
-	/* If we successfully set the value, we must query it back
+	/*
+	 * If we successfully set the value, we must query it back
 	 * Call Barring is a special case, since according to 22.088 2.2.1:
 	 * "The PLMN will ensure that only one of the barring programs is
 	 * active per basic service group. The activation of one specific
@@ -774,7 +749,7 @@ static void set_lock_callback(const struct ofono_error *error, void *data)
 	 * So we actually query all outgoing / incoming barrings depending
 	 * on what kind we set.
 	 */
-	set_query_next_lock(modem);
+	set_query_next_lock(cb);
 }
 
 static gboolean cb_lock_property_lookup(const char *property, const char *value,
@@ -808,10 +783,12 @@ static gboolean cb_lock_property_lookup(const char *property, const char *value,
 	} else if (!strcmp(property, "Incoming")) {
 		start = CB_INCOMING_START;
 		end = CB_INCOMING_END;
-	} else
+	} else {
 		return FALSE;
+	}
 
-	/* Gah, this is a special case.  If we're setting a barring to
+	/*
+	 * Gah, this is a special case.  If we're setting a barring to
 	 * disabled, then generate a disable all outgoing/incoming
 	 * request for a particular basic service
 	 */
@@ -844,8 +821,7 @@ static gboolean cb_lock_property_lookup(const char *property, const char *value,
 static DBusMessage *cb_set_property(DBusConnection *conn, DBusMessage *msg,
 					void *data)
 {
-	struct ofono_modem *modem = data;
-	struct call_barring_data *cb = modem->call_barring;
+	struct ofono_call_barring *cb = data;
 	DBusMessageIter iter;
 	DBusMessageIter var;
 	const char *name, *passwd = "";
@@ -854,7 +830,7 @@ static DBusMessage *cb_set_property(DBusConnection *conn, DBusMessage *msg,
 	int cls;
 	int mode;
 
-	if (cb->pending)
+	if (__ofono_call_barring_is_busy(cb) || __ofono_ussd_is_busy(cb->ussd))
 		return __ofono_error_busy(msg);
 
 	if (!dbus_message_iter_init(msg, &iter))
@@ -886,67 +862,61 @@ static DBusMessage *cb_set_property(DBusConnection *conn, DBusMessage *msg,
 			return __ofono_error_invalid_args(msg);
 
 		dbus_message_iter_get_basic(&iter, &passwd);
-		if (!is_valid_pin(passwd))
+		if (!is_valid_pin(passwd, PIN_TYPE_NET))
 			return __ofono_error_invalid_format(msg);
 	}
 
-	if (!cb->ops->set)
+	if (!cb->driver->set)
 		return __ofono_error_not_implemented(msg);
 
 	cb_set_query_bounds(cb, cb_locks[lock].fac, FALSE);
 
 	cb->pending = dbus_message_ref(msg);
-	cb->ops->set(modem, cb_locks[lock].fac, mode, passwd, cls,
-			set_lock_callback, modem);
+	cb->driver->set(cb, cb_locks[lock].fac, mode, passwd, cls,
+			set_lock_callback, cb);
 
 	return NULL;
 }
 
 static void disable_all_callback(const struct ofono_error *error, void *data)
 {
-	struct ofono_modem *modem = data;
-	struct call_barring_data *cb = modem->call_barring;
+	struct ofono_call_barring *cb = data;
 
 	if (error->type != OFONO_ERROR_TYPE_NO_ERROR) {
-		ofono_debug("Disabling all barring failed");
+		DBG("Disabling all barring failed");
 		__ofono_dbus_pending_reply(&cb->pending,
 					__ofono_error_failed(cb->pending));
 		return;
 	}
 
 	/* Assume if we have set, we have query */
-	set_query_next_lock(modem);
+	set_query_next_lock(cb);
 }
 
 static DBusMessage *cb_disable_all(DBusConnection *conn, DBusMessage *msg,
 					void *data, const char *fac)
 {
-	struct ofono_modem *modem = data;
-	struct call_barring_data *cb = modem->call_barring;
-	DBusMessageIter iter;
-	const char *passwd = "";
+	struct ofono_call_barring *cb = data;
+	const char *passwd;
 
-	if (cb->pending)
+	if (!cb->driver->set)
+		return __ofono_error_not_implemented(msg);
+
+	if (__ofono_call_barring_is_busy(cb) || __ofono_ussd_is_busy(cb->ussd))
 		return __ofono_error_busy(msg);
 
-	if (!dbus_message_iter_init(msg, &iter))
+	if (dbus_message_get_args(msg, NULL, DBUS_TYPE_STRING, &passwd,
+					DBUS_TYPE_INVALID) == FALSE)
 		return __ofono_error_invalid_args(msg);
 
-	if (dbus_message_iter_get_arg_type(&iter) != DBUS_TYPE_STRING)
-		return __ofono_error_invalid_args(msg);
-
-	dbus_message_iter_get_basic(&iter, &passwd);
-	if (!is_valid_pin(passwd))
+	if (!is_valid_pin(passwd, PIN_TYPE_NET))
 		return __ofono_error_invalid_format(msg);
 
-	if (!cb->ops->set)
-		return __ofono_error_not_implemented(msg);
-
 	cb_set_query_bounds(cb, fac, FALSE);
 
 	cb->pending = dbus_message_ref(msg);
-	cb->ops->set(modem, fac, 0, passwd,
-			BEARER_CLASS_DEFAULT, disable_all_callback, modem);
+	cb->driver->set(cb, fac, 0, passwd,
+			BEARER_CLASS_DEFAULT, disable_all_callback, cb);
 
 	return NULL;
 }
@@ -972,39 +942,30 @@ static DBusMessage *cb_disable_ag(DBusConnection *conn, DBusMessage *msg,
 static DBusMessage *cb_set_passwd(DBusConnection *conn, DBusMessage *msg,
 					void *data)
 {
-	struct ofono_modem *modem = data;
-	struct call_barring_data *cb = modem->call_barring;
-	DBusMessageIter iter;
-	const char *old_passwd, *new_passwd;
+	struct ofono_call_barring *cb = data;
+	const char *old_passwd;
+	const char *new_passwd;
 
-	if (cb->pending)
-		return __ofono_error_busy(msg);
-
-	if (!dbus_message_iter_init(msg, &iter))
-		return __ofono_error_invalid_args(msg);
-
-	if (dbus_message_iter_get_arg_type(&iter) != DBUS_TYPE_STRING)
-		return __ofono_error_invalid_args(msg);
-
-	dbus_message_iter_get_basic(&iter, &old_passwd);
-	if (!is_valid_pin(old_passwd))
-		return __ofono_error_invalid_format(msg);
-
-	dbus_message_iter_next(&iter);
-
-	if (dbus_message_iter_get_arg_type(&iter) != DBUS_TYPE_STRING)
-		return __ofono_error_invalid_args(msg);
-
-	dbus_message_iter_get_basic(&iter, &new_passwd);
-	if (!is_valid_pin(new_passwd))
-		return __ofono_error_invalid_format(msg);
-
-	if (!cb->ops->set_passwd)
+	if (!cb->driver->set_passwd)
 		return __ofono_error_not_implemented(msg);
 
+	if (__ofono_call_barring_is_busy(cb) || __ofono_ussd_is_busy(cb->ussd))
+		return __ofono_error_busy(msg);
+
+	if (dbus_message_get_args(msg, NULL, DBUS_TYPE_STRING, &old_passwd,
+					DBUS_TYPE_STRING, &new_passwd,
+					DBUS_TYPE_INVALID) == FALSE)
+		return __ofono_error_invalid_args(msg);
+
+	if (!is_valid_pin(old_passwd, PIN_TYPE_NET))
+		return __ofono_error_invalid_format(msg);
+
+	if (!is_valid_pin(new_passwd, PIN_TYPE_NET))
+		return __ofono_error_invalid_format(msg);
+
 	cb->pending = dbus_message_ref(msg);
-	cb->ops->set_passwd(modem, "AB", old_passwd, new_passwd,
-			cb_set_passwd_callback, modem);
+	cb->driver->set_passwd(cb, "AB", old_passwd, new_passwd,
+			cb_set_passwd_callback, cb);
 
 	return NULL;
 }
@@ -1034,92 +995,216 @@ static GDBusSignalTable cb_signals[] = {
 
 static void call_barring_incoming_enabled_notify(int idx, void *userdata)
 {
-	struct ofono_modem *modem = userdata;
+	struct ofono_call_barring *cb = userdata;
 	DBusConnection *conn = ofono_dbus_get_connection();
-	DBusMessage *signal;
+	const char *path = __ofono_atom_get_path(cb->atom);
 
-	signal = dbus_message_new_signal(modem->path,
-			CALL_BARRING_INTERFACE, "IncomingBarringInEffect");
-	if (!signal) {
-		ofono_error("Unable to allocate new %s.IncomingBarringInEffect"
-				" signal", CALL_BARRING_INTERFACE);
-		return;
-	}
-
-	g_dbus_send_message(conn, signal);
+	g_dbus_emit_signal(conn, path, OFONO_CALL_BARRING_INTERFACE,
+			"IncomingBarringInEffect", DBUS_TYPE_INVALID);
 }
 
 static void call_barring_outgoing_enabled_notify(int idx, void *userdata)
 {
-	struct ofono_modem *modem = userdata;
+	struct ofono_call_barring *cb = userdata;
 	DBusConnection *conn = ofono_dbus_get_connection();
+	const char *path = __ofono_atom_get_path(cb->atom);
 	DBusMessage *signal;
 
-	signal = dbus_message_new_signal(modem->path,
-			CALL_BARRING_INTERFACE, "OutgoingBarringInEffect");
+	signal = dbus_message_new_signal(path, OFONO_CALL_BARRING_INTERFACE,
+						"OutgoingBarringInEffect");
+
 	if (!signal) {
 		ofono_error("Unable to allocate new %s.OutgoingBarringInEffect"
-				" signal", CALL_BARRING_INTERFACE);
+				" signal", OFONO_CALL_BARRING_INTERFACE);
 		return;
 	}
 
 	g_dbus_send_message(conn, signal);
 }
 
-int ofono_call_barring_register(struct ofono_modem *modem,
-				struct ofono_call_barring_ops *ops)
+int ofono_call_barring_driver_register(const struct ofono_call_barring_driver *d)
 {
-	DBusConnection *conn = ofono_dbus_get_connection();
+	DBG("driver: %p, name: %s", d, d->name);
 
-	if (!modem || !ops)
-		return -1;
+	if (d->probe == NULL)
+		return -EINVAL;
 
-	modem->call_barring = call_barring_create();
-
-	if (!modem->call_barring)
-		return -1;
-
-	modem->call_barring->ops = ops;
-
-	if (!g_dbus_register_interface(conn, modem->path,
-					CALL_BARRING_INTERFACE,
-					cb_methods, cb_signals, NULL, modem,
-					call_barring_destroy)) {
-		ofono_error("Could not create %s interface",
-				CALL_BARRING_INTERFACE);
-		call_barring_destroy(modem);
-
-		return -1;
-	}
-
-	ofono_modem_add_interface(modem, CALL_BARRING_INTERFACE);
-
-	cb_register_ss_controls(modem);
-
-	ofono_mo_ss_register(modem, SS_MO_INCOMING_BARRING,
-			call_barring_incoming_enabled_notify, modem);
-	ofono_mo_ss_register(modem, SS_MO_OUTGOING_BARRING,
-			call_barring_outgoing_enabled_notify, modem);
+	g_drivers = g_slist_prepend(g_drivers, (void *) d);
 
 	return 0;
 }
 
-void ofono_call_barring_unregister(struct ofono_modem *modem)
+void ofono_call_barring_driver_unregister(const struct ofono_call_barring_driver *d)
 {
-	DBusConnection *conn = ofono_dbus_get_connection();
+	DBG("driver: %p, name: %s", d, d->name);
 
-	if (!modem->call_barring)
+	g_drivers = g_slist_remove(g_drivers, (void *) d);
+}
+
+static void call_barring_unregister(struct ofono_atom *atom)
+{
+	struct ofono_call_barring *cb = __ofono_atom_get_data(atom);
+	const char *path = __ofono_atom_get_path(cb->atom);
+	DBusConnection *conn = ofono_dbus_get_connection();
+	struct ofono_modem *modem = __ofono_atom_get_modem(cb->atom);
+
+	ofono_modem_remove_interface(modem, OFONO_CALL_BARRING_INTERFACE);
+	g_dbus_unregister_interface(conn, path, OFONO_CALL_BARRING_INTERFACE);
+
+	if (cb->ussd)
+		cb_unregister_ss_controls(cb);
+
+	if (cb->incoming_bar_watch)
+		__ofono_ssn_mo_watch_remove(cb->ssn, cb->incoming_bar_watch);
+	if (cb->outgoing_bar_watch)
+		__ofono_ssn_mt_watch_remove(cb->ssn, cb->outgoing_bar_watch);
+
+	if (cb->ssn_watch)
+		__ofono_modem_remove_atom_watch(modem, cb->ssn_watch);
+
+	if (cb->ussd_watch)
+		__ofono_modem_remove_atom_watch(modem, cb->ussd_watch);
+}
+
+static void call_barring_remove(struct ofono_atom *atom)
+{
+	struct ofono_call_barring *cb = __ofono_atom_get_data(atom);
+
+	DBG("atom: %p", atom);
+
+	if (cb == NULL)
 		return;
 
-	ofono_modem_remove_interface(modem, CALL_BARRING_INTERFACE);
-	g_dbus_unregister_interface(conn, modem->path, CALL_BARRING_INTERFACE);
+	if (cb->driver && cb->driver->remove)
+		cb->driver->remove(cb);
 
-	cb_unregister_ss_controls(modem);
+	g_free(cb);
+}
 
-	ofono_mo_ss_unregister(modem, SS_MO_INCOMING_BARRING,
-			call_barring_incoming_enabled_notify, modem);
-	ofono_mo_ss_unregister(modem, SS_MO_OUTGOING_BARRING,
-			call_barring_outgoing_enabled_notify, modem);
+struct ofono_call_barring *ofono_call_barring_create(struct ofono_modem *modem,
+							unsigned int vendor,
+							const char *driver,
+							void *data)
+{
+	struct ofono_call_barring *cb;
+	GSList *l;
 
-	modem->call_barring = NULL;
+	if (driver == NULL)
+		return NULL;
+
+	cb = g_try_new0(struct ofono_call_barring, 1);
+
+	if (cb == NULL)
+		return NULL;
+
+	cb->atom = __ofono_modem_add_atom(modem, OFONO_ATOM_TYPE_CALL_BARRING,
+						call_barring_remove, cb);
+
+	for (l = g_drivers; l; l = l->next) {
+		const struct ofono_call_barring_driver *drv = l->data;
+
+		if (g_strcmp0(drv->name, driver))
+			continue;
+
+		if (drv->probe(cb, vendor, data) < 0)
+			continue;
+
+		cb->driver = drv;
+		break;
+	}
+
+	return cb;
+}
+
+static void ssn_watch(struct ofono_atom *atom,
+			enum ofono_atom_watch_condition cond, void *data)
+{
+	struct ofono_call_barring *cb = data;
+
+	if (cond == OFONO_ATOM_WATCH_CONDITION_UNREGISTERED) {
+		cb->ssn = NULL;
+		cb->incoming_bar_watch = 0;
+		cb->outgoing_bar_watch = 0;
+		return;
+	}
+
+	cb->ssn = __ofono_atom_get_data(atom);
+
+	cb->incoming_bar_watch =
+		__ofono_ssn_mo_watch_add(cb->ssn, SS_MO_INCOMING_BARRING,
+				call_barring_incoming_enabled_notify, cb, NULL);
+
+	cb->outgoing_bar_watch =
+		__ofono_ssn_mo_watch_add(cb->ssn, SS_MO_OUTGOING_BARRING,
+				call_barring_outgoing_enabled_notify, cb, NULL);
+}
+
+static void ussd_watch(struct ofono_atom *atom,
+			enum ofono_atom_watch_condition cond, void *data)
+{
+	struct ofono_call_barring *cb = data;
+
+	if (cond == OFONO_ATOM_WATCH_CONDITION_UNREGISTERED) {
+		cb->ussd = NULL;
+		return;
+	}
+
+	cb->ussd = __ofono_atom_get_data(atom);
+	cb_register_ss_controls(cb);
+}
+
+void ofono_call_barring_register(struct ofono_call_barring *cb)
+{
+	DBusConnection *conn = ofono_dbus_get_connection();
+	const char *path = __ofono_atom_get_path(cb->atom);
+	struct ofono_modem *modem = __ofono_atom_get_modem(cb->atom);
+	struct ofono_atom *ssn_atom;
+	struct ofono_atom *ussd_atom;
+
+	if (!g_dbus_register_interface(conn, path,
+					OFONO_CALL_BARRING_INTERFACE,
+					cb_methods, cb_signals, NULL, cb,
+					NULL)) {
+		ofono_error("Could not create %s interface",
+				OFONO_CALL_BARRING_INTERFACE);
+
+		return;
+	}
+
+	ofono_modem_add_interface(modem, OFONO_CALL_BARRING_INTERFACE);
+
+	cb->ssn_watch = __ofono_modem_add_atom_watch(modem, OFONO_ATOM_TYPE_SSN,
+					ssn_watch, cb, NULL);
+
+	ssn_atom = __ofono_modem_find_atom(modem, OFONO_ATOM_TYPE_SSN);
+
+	if (ssn_atom && __ofono_atom_get_registered(ssn_atom))
+		ssn_watch(ssn_atom, OFONO_ATOM_WATCH_CONDITION_REGISTERED, cb);
+
+	cb->ussd_watch = __ofono_modem_add_atom_watch(modem,
+					OFONO_ATOM_TYPE_USSD,
+					ussd_watch, cb, NULL);
+
+	ussd_atom = __ofono_modem_find_atom(modem, OFONO_ATOM_TYPE_USSD);
+
+	if (ussd_atom && __ofono_atom_get_registered(ussd_atom))
+		ussd_watch(ussd_atom, OFONO_ATOM_WATCH_CONDITION_REGISTERED,
+				cb);
+
+	__ofono_atom_register(cb->atom, call_barring_unregister);
+}
+
+void ofono_call_barring_remove(struct ofono_call_barring *cb)
+{
+	__ofono_atom_free(cb->atom);
+}
+
+void ofono_call_barring_set_data(struct ofono_call_barring *cb, void *data)
+{
+	cb->driver_data = data;
+}
+
+void *ofono_call_barring_get_data(struct ofono_call_barring *cb)
+{
+	return cb->driver_data;
 }
